from .env_command import EnvCommand


class RemoveCommand(EnvCommand):
    """
    Removes a package from the project dependencies.

    remove
        { packages* : Packages that should be removed. }
        {--D|dev : Removes a package from the development dependencies. }
        {--dry-run : Outputs the operations but will not execute anything
                     (implicitly enables --verbose). }
    """

    help = """The <info>remove</info> command removes a package from the current
list of installed packages

<info>poetry remove</info>"""

    _loggers = ["poetry.repositories.pypi_repository"]

    def handle(self):
        from poetry.installation import Installer

        packages = self.argument("packages")
        is_dev = self.option("dev")

        original_content = self.poetry.file.read()
        content = self.poetry.file.read()
        poetry_content = content["tool"]["poetry"]
        section = "dependencies"
        if is_dev:
            section = "dev-dependencies"

        # Deleting entries
        requirements = {}
        for name in packages:
            found = False
            for key in poetry_content[section]:
                if key.lower() == name.lower():
                    found = True
                    requirements[key] = poetry_content[section][key]
                    break

            if not found:
                raise ValueError("Package {} not found".format(name))

        for key in requirements:
            del poetry_content[section][key]

        # Write the new content back
        self.poetry.file.write(content)

        # Update packages
        self.reset_poetry()

        installer = Installer(
            self.output,
            self.env,
            self.poetry.package,
            self.poetry.locker,
            self.poetry.pool,
        )

        installer.dry_run(self.option("dry-run"))
        installer.update(True)
        installer.whitelist(requirements)

        try:
            status = installer.run()
        except Exception:
            self.poetry.file.write(original_content)

            raise

        if status != 0 or self.option("dry-run"):
            # Revert changes
            if not self.option("dry-run"):
                self.error(
                    "\n"
                    "Removal failed, reverting pyproject.toml "
                    "to its original content."
                )

            self.poetry.file.write(original_content)

        return status



from poetry.utils._compat import Path
from poetry.utils.helpers import module_name

from .include import Include
from .package_include import PackageInclude


class Module:
    def __init__(self, name, directory=".", packages=None, includes=None):
        self._name = module_name(name)
        self._in_src = False
        self._is_package = False
        self._path = Path(directory)
        self._includes = []
        packages = packages or []
        includes = includes or []

        if not packages:
            # It must exist either as a .py file or a directory, but not both
            pkg_dir = Path(directory, self._name)
            py_file = Path(directory, self._name + ".py")
            if pkg_dir.is_dir() and py_file.is_file():
                raise ValueError("Both {} and {} exist".format(pkg_dir, py_file))
            elif pkg_dir.is_dir():
                packages = [{"include": str(pkg_dir.relative_to(self._path))}]
            elif py_file.is_file():
                packages = [{"include": str(py_file.relative_to(self._path))}]
            else:
                # Searching for a src module
                src = Path(directory, "src")
                src_pkg_dir = src / self._name
                src_py_file = src / (self._name + ".py")

                if src_pkg_dir.is_dir() and src_py_file.is_file():
                    raise ValueError("Both {} and {} exist".format(pkg_dir, py_file))
                elif src_pkg_dir.is_dir():
                    packages = [
                        {
                            "include": str(src_pkg_dir.relative_to(src)),
                            "from": str(src.relative_to(self._path)),
                        }
                    ]
                elif src_py_file.is_file():
                    packages = [
                        {
                            "include": str(src_py_file.relative_to(src)),
                            "from": str(src.relative_to(self._path)),
                        }
                    ]
                else:
                    raise ValueError("No file/folder found for package {}".format(name))

        for package in packages:
            self._includes.append(
                PackageInclude(self._path, package["include"], package.get("from"))
            )

        for include in includes:
            self._includes.append(Include(self._path, include))

    @property
    def name(self):  # type: () -> str
        return self._name

    @property
    def path(self):  # type: () -> Path
        return self._path

    @property
    def file(self):  # type: () -> Path
        if self._is_package:
            return self._path / "__init__.py"
        else:
            return self._path

    @property
    def includes(self):  # type: () -> List
        return self._includes

    def is_package(self):  # type: () -> bool
        return self._is_package

    def is_in_src(self):  # type: () -> bool
        return self._in_src



import logging

from cleo import Command as BaseCommand

from ..styles.poetry import PoetryStyle


class CommandFormatter(logging.Formatter):

    _colors = {
        "error": "fg=red",
        "warning": "fg=yellow",
        "debug": "debug",
        "info": "fg=blue",
    }

    def format(self, record):
        if not record.exc_info:
            level = record.levelname.lower()
            msg = record.msg

            if level in self._colors:
                msg = "<{}>{}</>".format(self._colors[level], msg)

            return msg

        return super(CommandFormatter, self).format(record)


class CommandHandler(logging.Handler):
    def __init__(self, command):
        self._command = command

        output = self._command.output
        level = logging.WARNING
        if output.is_debug():
            level = logging.DEBUG
        elif output.is_very_verbose() or output.is_verbose():
            level = logging.INFO

        super(CommandHandler, self).__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            level = record.levelname.lower()
            err = level in ("warning", "error", "exception", "critical")
            if err:
                self._command.output.write_error(msg, newline=True)
            else:
                self._command.line(msg)
        except Exception:
            self.handleError(record)


class Command(BaseCommand):

    _loggers = []

    @property
    def poetry(self):
        return self.get_application().poetry

    def reset_poetry(self):  # type: () -> None
        self.get_application().reset_poetry()

    def run(self, i, o):  # type: () -> int
        """
        Initialize command.
        """
        self.input = i
        self.output = PoetryStyle(i, o)

        for logger in self._loggers:
            self.register_logger(logging.getLogger(logger))

        return super(BaseCommand, self).run(i, o)

    def register_logger(self, logger):
        """
        Register a new logger.
        """
        handler = CommandHandler(self)
        handler.setFormatter(CommandFormatter())
        logger.handlers = [handler]
        logger.propagate = False

        output = self.output
        level = logging.WARNING
        if output.is_debug():
            level = logging.DEBUG
        elif output.is_very_verbose() or output.is_verbose():
            level = logging.INFO

        logger.setLevel(level)



from poetry.semver import parse_constraint
from poetry.semver import Version
from poetry.semver import VersionUnion

PYTHON_VERSION = [
    "2.7.*",
    "3.0.*",
    "3.1.*",
    "3.2.*",
    "3.3.*",
    "3.4.*",
    "3.5.*",
    "3.6.*",
    "3.7.*",
    "3.8.*",
]


def format_python_constraint(constraint):
    """
    This helper will help in transforming
    disjunctive constraint into proper constraint.
    """
    if isinstance(constraint, Version) and constraint.precision < 3:
        # Transform 3.6 or 3
        if constraint.precision == 2:
            # 3.6
            constraint = parse_constraint(
                "~{}.{}".format(constraint.major, constraint.minor)
            )
        else:
            constraint = parse_constraint("^{}.0".format(constraint.major))

    if not isinstance(constraint, VersionUnion):
        return str(constraint)

    formatted = []
    accepted = []

    for version in PYTHON_VERSION:
        version_constraint = parse_constraint(version)
        matches = constraint.allows_any(version_constraint)
        if not matches:
            formatted.append("!=" + version)
        else:
            accepted.append(version)

    # Checking lower bound
    low = accepted[0]

    formatted.insert(0, ">=" + ".".join(low.split(".")[:2]))

    return ", ".join(formatted)



from .command import Command


class EnvCommand(Command):
    def __init__(self):
        self._env = None

        super(EnvCommand, self).__init__()

    def initialize(self, i, o):
        from poetry.utils.env import Env

        super(EnvCommand, self).initialize(i, o)

        self._env = Env.create_venv(
            o, self.poetry.package.name, cwd=self.poetry.file.parent
        )

        if self._env.is_venv() and o.is_verbose():
            o.writeln("Using virtualenv: <comment>{}</>".format(self._env.path))

    @property
    def env(self):
        return self._env



from .command import Command


class SearchCommand(Command):
    """
    Searches for packages on remote repositories.

    search
        { tokens* : The tokens to search for. }
        { --N|only-name : Search only in name. }
    """

    def handle(self):
        from poetry.repositories.pypi_repository import PyPiRepository

        flags = PyPiRepository.SEARCH_FULLTEXT
        if self.option("only-name"):
            flags = PyPiRepository.SEARCH_FULLTEXT

        results = PyPiRepository().search(self.argument("tokens"), flags)

        for result in results:
            self.line("")
            name = "<info>{}</>".format(result.name)

            name += " (<comment>{}</>)".format(result.version)

            self.line(name)

            if result.description:
                self.line(" {}".format(result.description))



import subprocess
import warnings

from poetry.utils._compat import Path

from .git import Git


def get_vcs(directory):  # type: (Path) -> Git
    directory = directory.resolve()

    for p in [directory] + list(directory.parents):
        if (p / ".git").is_dir():
            try:
                return Git(p)
            except (subprocess.CalledProcessError, OSError):
                # Either git could not be found or does not exist
                warnings.warn(
                    "git executable could not be found", category=RuntimeWarning
                )

                return



import sys

from cleo.inputs import ArgvInput


class RawArgvInput(ArgvInput):
    def parse(self):
        self._parsed = self._tokens
        while True:
            try:
                token = self._parsed.pop(0)
            except IndexError:
                break

            self.parse_argument(token)



import os
import sys

from ..command import Command


class DebugInfoCommand(Command):
    """
    Shows debug information.

    debug:info
    """

    def handle(self):
        from ....utils.env import Env

        poetry = self.poetry
        package = poetry.package
        env = Env.get(cwd=poetry.file.parent)

        poetry_python_version = ".".join(str(s) for s in sys.version_info[:3])

        self.output.title("Poetry")
        self.output.listing(
            [
                "<info>Version</info>: <comment>{}</>".format(poetry.VERSION),
                "<info>Python</info>:  <comment>{}</>".format(poetry_python_version),
            ]
        )

        self.line("")

        env_python_version = ".".join(str(s) for s in env.version_info[:3])
        self.output.title("Virtualenv")
        self.output.listing(
            [
                "<info>Python</info>:         <comment>{}</>".format(
                    env_python_version
                ),
                "<info>Implementation</info>: <comment>{}</>".format(
                    env.python_implementation
                ),
                "<info>Path</info>:           <comment>{}</>".format(
                    env.path if env.is_venv() else "NA"
                ),
            ]
        )

        self.line("")

        self.output.title("System")
        self.output.listing(
            [
                "<info>Platform</info>: <comment>{}</>".format(sys.platform),
                "<info>OS</info>:       <comment>{}</>".format(os.name),
                "<info>Python</info>:   <comment>{}</>".format(env.base),
            ]
        )

        self.line("")



class CompatibilityError(Exception):
    def __init__(self, *constraints):
        self._constraints = list(constraints)

    @property
    def constraints(self):
        return self._constraints


class SolverProblemError(Exception):
    def __init__(self, error):
        self._error = error

        super(SolverProblemError, self).__init__(str(error))

    @property
    def error(self):
        return self._error
